import numpy as np
import torch
from torch.utils.data import Dataset

np.random.seed(12345)

class Word2vecDataset(Dataset):
    def __init__(self, inputFileName, side_num = 0, neg_num = 5, sentences_count = 1000000):
        # self.data = data
        # self.window_size = window_size
        self.sentences_count = sentences_count
        self.neg_num = neg_num
        self.side_num = side_num
        self.input_file = open(inputFileName, encoding="utf8")
        self.sample_size=64*1000
        self.data = []
        self.weights=[]
        self.choices = []
        self.idx = 0

    def __len__(self):
        return self.sentences_count

    def __getitem__(self, idx):
        # no shuffle
        while True:
            if self.idx%self.sample_size==0:
                self.data = []
                self.idx = 0
                for i in range(self.sample_size):
                    line = self.input_file.readline().strip('\n')
                    if not line:
                        self.input_file.seek(0, 0)
                        line = self.input_file.readline().strip('\n')
                    if len(line) > 1 and len(line.split(' ')) == self.neg_num+3:
                        self.data.append(line)
                self.weights=[float(x.split(' ')[-1]) for x in self.data]
                self.choices = np.random.choice(a=self.data, size=self.sample_size, replace=True, p=np.array(self.weights)/sum(self.weights))

            # print(self.data)
            line = self.choices[self.idx%self.sample_size]
            words = [x for x in line.split(' ')[:-1]]
            assert len(words) == self.neg_num+2, str(words)
            assert len(words[0].split('_')) == self.side_num+1, str(words)
            self.idx += 1
            return [words[0],words[1],words[2:]]


    @staticmethod
    def collate(batches):
        all_u = [batch[0] for batch in batches if len(batch) > 0]
        all_u = [list(map(int,x.split('_'))) for x in all_u]
        all_v = [batch[1] for batch in batches if len(batch) > 0]
        all_v = [list(map(int,x.split('_'))) for x in all_v]
        all_neg_v = [batch[2] for batch in batches if len(batch) > 0]
        all_neg_v = [[list(map(int,x.split('_'))) for x in xx] for xx in all_neg_v]
        return torch.LongTensor(all_u), torch.LongTensor(all_v), torch.LongTensor(all_neg_v)



import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn import init

"""
    u_embedding: Embedding for center word.
    v_embedding: Embedding for neighbor words.
"""


class SkipGramModel(nn.Module):

    def __init__(self, emb_size, emb_dimension):
        super(SkipGramModel, self).__init__()
        self.emb_size = emb_size
        self.emb_dimension = emb_dimension
        self.u_embeddings = []
        self.v_embeddings = []
        initrange = 1.0 / self.emb_dimension / len(emb_size)
        self.u_embeddings = nn.ModuleList([nn.Embedding(emb_size[i], emb_dimension, sparse=True) for i in range(len(emb_size))])
        self.v_embeddings = nn.ModuleList([nn.Embedding(emb_size[i], emb_dimension, sparse=True) for i in range(len(emb_size))])

        for i in range(len(emb_size)):
            init.uniform_(self.u_embeddings[i].weight.data, -initrange, initrange)
            init.constant_(self.v_embeddings[i].weight.data, 0)

    def forward(self, pos_u, pos_v, neg_v):
        emb_u = []
        for i in range(len(self.emb_size)):
            emb_u.append(self.u_embeddings[i](pos_u[:,i]))
        emb_u = torch.stack(emb_u, dim=1)

        score_all = []
        neg_score_all = []

        for i in range(len(self.emb_size)):
            emb_v = self.v_embeddings[i](pos_v[:,i])
            emb_neg_v = self.v_embeddings[i](neg_v[:,:,i])

            score = torch.bmm(emb_u,emb_v.unsqueeze(2))
            score = torch.clamp(score, max=10, min=-10)
            score_all.append(-torch.sum(F.logsigmoid(-score), dim=1).squeeze())

            neg_score = torch.bmm(emb_neg_v, torch.transpose(emb_u, 1, 2))
            neg_score = torch.clamp(neg_score, max=10, min=-10)
            neg_score_all.append(-torch.sum(F.logsigmoid(-neg_score), dim=1).squeeze())

        score_all = torch.stack(score_all, dim=1)
        neg_score_all = torch.stack(neg_score_all, dim=1)

        return torch.mean(score_all + neg_score_all)

    def save_embedding(self, file_name):
        for i in range(len(self.emb_size)):
            embedding = self.u_embeddings[i].weight.cpu().data.numpy()
            with open(file_name+'_'+str(i), 'w') as f:
                for j in range(len(embedding)):
                    f.write('%s %s\n' % (j, ' '.join(list(map(str,embedding[j])))))



import torch
import torch.optim as optim
from torch.utils.data import DataLoader, sampler
from tqdm import tqdm
from word2vec.data_reader import Word2vecDataset
from word2vec.model import SkipGramModel


class Word2VecTrainer:
    def __init__(self, input_file, output_file, side_num=1, neg_num=5, sentences_count=100000, emb_size=[1000000,],
                 emb_dimension=100, batch_size=32, iterations=3,
                 initial_lr=0.001):

        # self.data = DataReader(input_file, min_count)
        dataset = Word2vecDataset(input_file, side_num=side_num , neg_num=neg_num , sentences_count=sentences_count)
        self.dataloader = DataLoader(dataset, batch_size=batch_size,
                                     shuffle=False, num_workers=1, collate_fn=dataset.collate)

        self.output_file_name = output_file
        self.emb_size = emb_size
        self.emb_dimension = emb_dimension
        self.batch_size = batch_size
        self.iterations = iterations
        self.initial_lr = initial_lr
        self.skip_gram_model = SkipGramModel(self.emb_size, self.emb_dimension)

        self.use_cuda = torch.cuda.is_available()
        self.device = torch.device("cuda" if self.use_cuda else "cpu")
        if self.use_cuda:
            self.skip_gram_model.cuda()

    def train(self):

        for iteration in range(self.iterations):

            print("\n\n\nIteration: " + str(iteration + 1))
            optimizer = optim.SparseAdam(self.skip_gram_model.parameters(), lr=self.initial_lr)
            scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, len(self.dataloader))

            running_loss = 0.0
            for i, sample_batched in enumerate(tqdm(self.dataloader)):

                if len(sample_batched[0]) > 1:
                    pos_u = sample_batched[0].to(self.device)
                    pos_v = sample_batched[1].to(self.device)
                    neg_v = sample_batched[2].to(self.device)

                    scheduler.step()
                    optimizer.zero_grad()
                    loss = self.skip_gram_model.forward(pos_u, pos_v, neg_v)
                    loss.backward()
                    optimizer.step()

                    running_loss = running_loss * 0.9 + loss.item() * 0.1
                    if i<=1 or i % 500 == 0:
                        print(" Loss: " + str(running_loss))

            self.skip_gram_model.save_embedding(self.output_file_name)




from setuptools import setup, find_packages

setup(
    name='word2vec-SISG',
    version='1.0.0',
    description='<Billion-scale Recommendation with Heterogeneous Side Information at Taobao> PyTorch implementation',
    packages=find_packages(include=['word2vec']),
    install_requires=[],
)

